package com.fjhx.service.impl;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.io.FileUtil;
import cn.hutool.core.lang.Assert;
import cn.hutool.core.thread.ThreadUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.baomidou.mybatisplus.core.toolkit.ObjectUtils;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.fjhx.config.TaskPoolConfig;
import com.fjhx.config.exception.EmailEngineException;
import com.fjhx.config.redis.RedisCache;
import com.fjhx.constants.RedisConstant;
import com.fjhx.constants.SendConstant;
import com.fjhx.entity.EmailInfo;
import com.fjhx.entity.EmailMessage;
import com.fjhx.entity.EmailMessageAttachment;
import com.fjhx.entity.EmailMessageSend;
import com.fjhx.enums.SendEventEnum;
import com.fjhx.service.*;
import com.fjhx.utils.EmailEngineUtil;
import com.fjhx.utils.RetryUtil;
import com.fjhx.vo.*;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Executor;

@Slf4j
@Service
public class AccountServiceImpl implements IAccountService {

    @Autowired
    private IEmailInfoService emailInfoService;

    @Autowired
    private IEmailMessageService emailMessageService;

    @Autowired
    private IEmailMessageSendService emailMessageSendService;

    @Autowired
    private IEmailMessageAttachmentService emailMessageAttachmentService;

    @Qualifier(TaskPoolConfig.taskExecutor)
    @Autowired
    private Executor executor;

    @Transactional(rollbackFor = Exception.class)
    @Override
    public EmailInfo binding(BindingVo bindingVo) {

        EmailInfo emailInfo;

        try {
            String email = bindingVo.getEmail();

            emailInfo = emailInfoService.getOne(Wrappers.<EmailInfo>lambdaQuery().eq(EmailInfo::getEmail, email));
            // 如果存在，直接返回邮箱信息
            if (emailInfo != null) {
                return emailInfo;
            }
            // 保存账号信息到数据库
            emailInfo = this.saveEmailInfo(bindingVo);
            // 添加账号
            EmailEngineUtil.createAccount(bindingVo);
            // redis添加同步进度
            this.progressInitialization(bindingVo);

        } catch (Exception e) {
            EmailEngineUtil.deleteAccount(bindingVo.getEmail());
            throw e;
        }

        // 异步遍历文件夹下的所有邮件
        this.asyncReadEmail(emailInfo.getId(), bindingVo, 19);
        return emailInfo;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void listener(ListenerVo listenerVo) {

        log.info("监听到新消息：{}", JSON.toJSONString(listenerVo));

        switch (SendEventEnum.get(listenerVo.getEvent())) {
            case MESSAGE_NEW:
                this.handleMessageNewEvent(listenerVo);
                break;
            case MESSAGE_DELETED:
                this.handleMessageDeletedEvent(listenerVo);
                break;
            case MESSAGE_UPDATED:
                this.handleMessageUpdatedEvent(listenerVo);
                break;
            default:
                log.error("监听到未知事件：{}", JSONObject.toJSONString(listenerVo));
        }

    }

    @Override
    public void submit(SubmitVo submitVo) {
        EmailEngineUtil.submit(submitVo);
    }

    /**
     * 添加邮件信息
     */
    private EmailInfo saveEmailInfo(BindingVo bindingVo) {
        EmailInfo emailInfo = BeanUtil.copyProperties(bindingVo, EmailInfo.class);
        emailInfoService.save(emailInfo);
        return emailInfo;
    }

    /**
     * 同步邮件进度初始化
     */
    private void progressInitialization(BindingVo bindingVo) {
        String email = bindingVo.getEmail();
        int pages = bindingVo.getPages();

        MessageVo messageVo = EmailEngineUtil.getMessageList(email, bindingVo.getPath(), 1);

        ProgressVo progressVo = new ProgressVo();
        progressVo.setEmail(email);
        progressVo.setCompleteMessageCount(0);
        progressVo.setPercentage(0);
        progressVo.setTotalMessageCount(messageVo.getTotal() > pages * 10 ? pages * 10 : messageVo.getTotal());
        RedisCache.set(RedisConstant.PROGRESS_KEY + email, progressVo);
    }

    /**
     * 异步读取文件
     */
    private void asyncReadEmail(Long emailInfoId, BindingVo bindingVo, Integer pageFlag) {
        String email = bindingVo.getEmail();
        int pages = bindingVo.getPages();
        String mailbox = bindingVo.getPath();

        executor.execute(() -> {
            ThreadUtil.sleep(3000);
            int page = pageFlag;
            try {

                // 记录上次消息id,防止循环查询分页过程中接收到了新邮件，出现邮件重复问题
                List<String> lastMessageIdList = new ArrayList<>();

                while (page < pages) {

                    // 分页获取文件夹邮件
                    MessageVo result = EmailEngineUtil.getMessageList(email, mailbox, page);

                    List<MessageVo.MessagesDTO> messagesDTOList = result.getMessages();

                    if (messagesDTOList.size() > 0) {

                        // 批量保存邮件信息
                        List<EmailMessage> emailMessageList = this.saveBatchMessage(emailInfoId, email, lastMessageIdList, messagesDTOList);

                        // 更新同步邮件进度
                        this.synchronizationProgress(email, messagesDTOList);

                        // 记录上次保存的邮件id
                        lastMessageIdList.clear();
                        for (EmailMessage emailMessage : emailMessageList) {
                            lastMessageIdList.add(emailMessage.getMessageId());
                        }

                    }
                    page++;

                    ThreadUtil.sleep(3000);
                }

            } catch (EmailEngineException e) {
                log.error("下载失败，等待5分钟");
                ThreadUtil.sleep(1000 * 60 * 5);

                asyncReadEmail(emailInfoId, bindingVo, page);
            } catch (Exception e) {
                log.error("未知原因同步文件夹失败：{}", email, e);
            }

        });

    }

    /**
     * 批量保存邮件信息
     */
    private List<EmailMessage> saveBatchMessage(Long emailInfoId, String email, List<String> lastMessageId, List<MessageVo.MessagesDTO> messageList) {
        List<EmailMessage> emailMessageList = new ArrayList<>();
        List<EmailMessageSend> emailMessageSendList = new ArrayList<>();
        List<EmailMessageAttachment> emailMessageAttachmentList = new ArrayList<>();

        for (MessageVo.MessagesDTO messagesDTO : messageList) {

            // 已添加过的邮件则跳过
            if (lastMessageId.contains(messagesDTO.getId())) {
                continue;
            }

            // 生成邮件实体
            EmailMessage emailMessage = this.createMessage(emailInfoId, email, messagesDTO);
            emailMessageList.add(emailMessage);

            // 添加推送信息
            this.addMessageSend(emailMessageSendList, messagesDTO, emailMessage.getId());

            // 下载并添加附件信息
            this.addMessageAttachment(email, emailMessageAttachmentList, messagesDTO, emailMessage.getId());
        }

        // 保存邮件
        emailMessageService.saveBatch(emailMessageList);
        // 保存推送人信息
        emailMessageSendService.saveBatch(emailMessageSendList);
        // 保存附件信息
        emailMessageAttachmentService.saveBatch(emailMessageAttachmentList);

        return emailMessageList;
    }

    /**
     * 生成邮件实体
     */
    private EmailMessage createMessage(Long emailInfoId, String email, MessageVo.MessagesDTO messagesDTO) {

        EmailMessage emailMessage = new EmailMessage();
        emailMessage.setId(IdWorker.getId());
        emailMessage.setMessageId(messagesDTO.getId());
        emailMessage.setEmailInfoId(emailInfoId);
        emailMessage.setEmail(email);
        emailMessage.setUnseen(messagesDTO.getUnseen());
        emailMessage.setFlagged(messagesDTO.getFlagged());
        emailMessage.setSubject(messagesDTO.getSubject());
        emailMessage.setFromDate(messagesDTO.getDate());

        // 发件人
        MessageVo.MessagesDTO.ToDTO from = messagesDTO.getFrom();
        if (ObjectUtils.isNotEmpty(from)) {
            emailMessage.setFromName(from.getName());
            emailMessage.setFromAddress(from.getAddress());
        }

        // 正文内容
        MessageDetailVo messageDetail = EmailEngineUtil.getMessageDetail(email, messagesDTO.getId());
        MessageDetailVo.TextDTO text = messageDetail.getText();
        if (ObjectUtils.isNotEmpty(text)) {
            emailMessage.setHtmlContent(text.getHtml());
        }

        return emailMessage;
    }

    /**
     * 添加推送信息
     */
    private void addMessageSend(List<EmailMessageSend> emailMessageSendList, MessageVo.MessagesDTO messagesDTO, Long emailMessageId) {
        // 收件人
        List<MessageVo.MessagesDTO.ToDTO> to = messagesDTO.getTo();
        if (ObjectUtils.isNotEmpty(to)) {
            for (MessageVo.MessagesDTO.ToDTO toDTO : to) {
                EmailMessageSend emailMessageSend = new EmailMessageSend();
                emailMessageSend.setEmailMessageId(emailMessageId);
                emailMessageSend.setMessageId(messagesDTO.getId());
                emailMessageSend.setType(SendConstant.TO);
                emailMessageSend.setName(toDTO.getName());
                emailMessageSend.setAddress(toDTO.getAddress());
                emailMessageSendList.add(emailMessageSend);
            }
        }

        // 抄送人
        List<MessageVo.MessagesDTO.ToDTO> cc = messagesDTO.getCc();
        if (ObjectUtils.isNotEmpty(cc)) {
            for (MessageVo.MessagesDTO.ToDTO toDTO : cc) {
                EmailMessageSend emailMessageSend = new EmailMessageSend();
                emailMessageSend.setEmailMessageId(emailMessageId);
                emailMessageSend.setMessageId(messagesDTO.getId());
                emailMessageSend.setType(SendConstant.CC);
                emailMessageSend.setName(toDTO.getName());
                emailMessageSend.setAddress(toDTO.getAddress());
                emailMessageSendList.add(emailMessageSend);
            }
        }

        // 密送人
        List<MessageVo.MessagesDTO.ToDTO> bcc = messagesDTO.getBcc();
        if (ObjectUtils.isNotEmpty(bcc)) {
            for (MessageVo.MessagesDTO.ToDTO toDTO : cc) {
                EmailMessageSend emailMessageSend = new EmailMessageSend();
                emailMessageSend.setEmailMessageId(emailMessageId);
                emailMessageSend.setMessageId(messagesDTO.getId());
                emailMessageSend.setType(SendConstant.BCC);
                emailMessageSend.setName(toDTO.getName());
                emailMessageSend.setAddress(toDTO.getAddress());
                emailMessageSendList.add(emailMessageSend);
            }
        }
    }

    /**
     * 下载并添加附件信息
     */
    private void addMessageAttachment(String email, List<EmailMessageAttachment> emailMessageAttachmentList,
                                      MessageVo.MessagesDTO messagesDTO, Long emailMessageId) {

        List<MessageVo.MessagesDTO.AttachmentsDTO> attachments = messagesDTO.getAttachments();
        if (ObjectUtils.isNotEmpty(attachments)) {
            for (MessageVo.MessagesDTO.AttachmentsDTO attachment : attachments) {
                String attachmentId = attachment.getId();

                String newFileName = attachmentId + "." + FileUtil.getSuffix(attachment.getFilename());

                EmailMessageAttachment emailMessageAttachment = new EmailMessageAttachment();
                emailMessageAttachment.setEmailMessageId(emailMessageId);
                emailMessageAttachment.setMessageId(messagesDTO.getId());
                emailMessageAttachment.setName(attachment.getFilename());
                emailMessageAttachment.setPath(email + "\\" + newFileName);
                emailMessageAttachment.setAttachmentId(attachmentId);
                emailMessageAttachment.setSize(attachment.getEncodedSize());
                emailMessageAttachmentList.add(emailMessageAttachment);

                // 下载附件
                try {
                    RetryUtil.execute(() -> EmailEngineUtil.downloadAttachment(email, attachmentId, newFileName));
                } catch (Exception e) {
                    log.error("下载附件失败：email:{}，attachmentId:{}", email, attachmentId, e);

                    RedisCache.set(RedisConstant.DOWNLOAD_FAIL_FILE_KEY + email + ":" + attachmentId, emailMessageAttachment);
                }

            }
        }
    }

    /**
     * 同步邮件进度
     */
    private void synchronizationProgress(String email, List<MessageVo.MessagesDTO> messagesDTOList) {
        ProgressVo progressVo = RedisCache.get(RedisConstant.PROGRESS_KEY + email);
        int completeMessageCount = progressVo.getCompleteMessageCount() + messagesDTOList.size();
        progressVo.setCompleteMessageCount(completeMessageCount);
        progressVo.setPercentage(100 * completeMessageCount / progressVo.getTotalMessageCount());
        RedisCache.set(RedisConstant.PROGRESS_KEY + email, progressVo);
    }

    /**
     * 处理收到新邮件事件
     */
    private void handleMessageNewEvent(ListenerVo listenerVo) {
        // 查询邮箱
        String account = listenerVo.getAccount();

        EmailInfo emailInfo;
        try {
            emailInfo = RetryUtil.execute(() -> {
                EmailInfo tempEmailInfo = emailInfoService.getOne(Wrappers.<EmailInfo>lambdaQuery()
                        .eq(EmailInfo::getEmail, account));
                Assert.notNull(tempEmailInfo, "未找到邮箱信息");
                return tempEmailInfo;
            });
        } catch (Exception e) {
            log.error("查找邮箱失败", e);
            return;
        }

        // 获取附件详情
        MessageVo.MessagesDTO messagesDTO = listenerVo.getData().toJavaObject(MessageVo.MessagesDTO.class);

        // 保存邮件
        EmailMessage emailMessage = this.createMessage(emailInfo.getId(), emailInfo.getEmail(), messagesDTO);
        emailMessageService.save(emailMessage);

        // 保存推送人信息
        List<EmailMessageSend> emailMessageSendList = new ArrayList<>();
        this.addMessageSend(emailMessageSendList, messagesDTO, emailMessage.getId());
        emailMessageSendService.saveBatch(emailMessageSendList);

        // 保存附件信息
        List<EmailMessageAttachment> emailMessageAttachmentList = new ArrayList<>();
        this.addMessageAttachment(account, emailMessageAttachmentList, messagesDTO, emailMessage.getId());
        emailMessageAttachmentService.saveBatch(emailMessageAttachmentList);
    }

    /**
     * 处理删除邮件事件
     */
    private void handleMessageDeletedEvent(ListenerVo listenerVo) {
        MessageDeletedEventVo vo = listenerVo.getData().toJavaObject(MessageDeletedEventVo.class);
        emailMessageService.remove(
                Wrappers.<EmailMessage>lambdaQuery().eq(EmailMessage::getMessageId, vo.getId()));
    }

    /**
     * 邮件flag被改变事件
     */
    private void handleMessageUpdatedEvent(ListenerVo listenerVo) {
        log.error("邮件flag改变：{}", JSON.toJSONString(listenerVo));
    }

}
